#[doc(hidden)]
#[macro_export]
macro_rules! __array_op {
    (gen[$len:expr] |$i:pat_param| $val:expr) => {{
        let mut out = std::mem::MaybeUninit::uninit_array::<$len>();
        let mut i = 0;
        while i < $len {
            out[i] = std::mem::MaybeUninit::new(match i {
                $i => $val,
            });
            i += 1;
        }
        #[allow(unused_unsafe)]
        unsafe {
            std::mem::MaybeUninit::array_assume_init(out)
        }
    }};
    (map[$len:expr, $src:expr] |$i:pat_param, $s:pat_param| $val:expr) => {{
        $crate::util::array_op!(
            gen[$len]
                | i
                | match i {
                    $i => match $src[i] {
                        $s => $val,
                    },
                }
        )
    }};
}

pub use __array_op as array_op;

#[inline]
pub const fn cast_u8_u32<const N: usize>(arr: [u8; N]) -> [u32; N] {
    array_op!(map[N, arr] |_, v| v as u32)
}

#[doc(hidden)]
#[macro_export]
macro_rules! __defer_impl {
    (
        => $impl:ident;

        $( fn $name:ident($( $pname:ident: $pty:ty ),*) -> $rty:ty; )*
    ) => {
        $(
            #[inline(always)]
            fn $name($( $pname: $pty ),*) -> $rty {
                <$impl>::$name($( $pname ),*)
            }
        )*
    };
}

pub use __defer_impl as defer_impl;

#[inline(always)]
#[cold]
pub const fn cold() {}

#[inline]
pub const fn likely(b: bool) -> bool {
    if !b {
        cold()
    }
    b
}

#[inline]
pub const fn unlikely(b: bool) -> bool {
    if b {
        cold()
    }
    b
}

/// Like transmute, but implemented via a union so that we can use it in situations where
/// transmute's "safety" restrictions are too strict and uninformed (i.e. we can prove it is safe).
#[inline(always)]
pub const unsafe fn cast<A, B>(a: A) -> B {
    union Cast<A, B> {
        a: core::mem::ManuallyDrop<A>,
        b: core::mem::ManuallyDrop<B>,
    }

    core::mem::ManuallyDrop::into_inner(
        Cast {
            a: core::mem::ManuallyDrop::new(a),
        }
        .b,
    )
}

#[doc(hidden)]
#[macro_export]
macro_rules! __subst {
        ([$( $ignore:tt )*], [$( $use:tt )*]) => {
            $( $use )*
        };
    }

pub use __subst as subst;

#[inline(always)]
pub const fn align_down_to<const N: usize>(n: usize) -> usize {
    let shift = match N.checked_ilog2() {
        Some(x) => x,
        None => 0,
    };
    return n >> shift << shift;
}

#[inline(always)]
pub const fn align_up_to<const N: usize>(n: usize) -> usize {
    let shift = match N.checked_ilog2() {
        Some(x) => x,
        None => 0,
    };
    return (n + (N - 1)) >> shift << shift;
}

#[inline(always)]
pub unsafe fn alloc_aligned_box<T, const ALIGN: usize>() -> Box<T> {
    let ptr = alloc_aligned(
        std::mem::size_of::<T>(),
        std::cmp::max(std::mem::align_of::<T>(), ALIGN),
    );
    Box::from_raw(ptr as *mut T)
}

#[inline(always)]
pub unsafe fn alloc_aligned_box_slice<T, const ALIGN: usize>(len: usize) -> Box<[T]> {
    let size = std::cmp::max(std::mem::size_of::<T>(), std::mem::align_of::<T>());
    let ptr = alloc_aligned(size * len, std::cmp::max(std::mem::align_of::<T>(), ALIGN));
    Box::from_raw(std::ptr::slice_from_raw_parts_mut(ptr as *mut T, len))
}

#[inline(always)]
pub unsafe fn alloc_aligned(len: usize, align: usize) -> *mut u8 {
    let layout = std::alloc::Layout::from_size_align_unchecked(len, align);
    let ptr = std::alloc::alloc(layout);
    if ptr.is_null() {
        std::alloc::handle_alloc_error(layout);
    }
    ptr
}

#[macro_export]
macro_rules! __util__unroll {
    (let [$( $x:ident ),+] => |$y:pat_param| $expr:expr) => {
        $crate::util::unroll!(let [$( $x: ($x) ),+] => |$y| $expr);
    };
    (let [$( $id:ident: ($( $x:expr ),+) ),+] => |$( $y:pat_param ),+| $($expr:tt)+) => {
        $crate::util::unroll!(@ [$] [$( $id: ($( $x ),+) ),+] => |$( $y ),+| $($expr)+)
    };
    (@ [$dollar:tt] [$( $id:ident: ($( $x:expr ),+) ),+] => |$( $y:pat_param ),+| $($expr:tt)+) => {
        macro_rules! __unrolled {
            ($dollar id:ident, $dollar ( $dollar z:tt ),+) => {
                #[allow(unused_parens)]
                let ($( $y ),+) = ($dollar ( $dollar z ),+);
                let $dollar id = $($expr)+;
            };
        }
        $( __unrolled!($id, $( $x ),+); )+
        //$(
        //    let ($( $y ),+) = ($( $x ),+);
        //    $expr;
        //)+
    };
    ([$( ($( $x:expr ),+) ),+] => |$( $y:pat_param ),+| $($expr:tt)+) => {
        $crate::util::unroll!(@ [$] [$( ($( $x ),+) ),+] => |$( $y ),+| $($expr)+)
    };
    (@ [$dollar:tt] [$( ($( $x:expr ),+) ),+] => |$( $y:pat_param ),+| $($expr:tt)+) => {
        macro_rules! __unrolled {
            ($dollar ( $dollar z:tt ),+) => {
                #[allow(unused_parens)]
                let ($( $y ),+) = ($dollar ( $dollar z ),+);
                $($expr)+;
            };
        }
        $( __unrolled!($( $x ),+); )+
        //$(
        //    let ($( $y ),+) = ($( $x ),+);
        //    $expr;
        //)+
    };
}

pub use __util__unroll as unroll;

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    pub fn test_align_down_to() {
        assert_eq!(align_down_to::<8>(8), 8);
        assert_eq!(align_down_to::<16>(8), 0);
        assert_eq!(align_down_to::<16>(16), 16);
        assert_eq!(align_down_to::<16>(15), 0);
        assert_eq!(align_down_to::<16>(17), 16);
    }

    #[test]
    pub fn test_array_op_gen() {
        assert_eq!(array_op!(gen[4] | i | i), [0, 1, 2, 3]);
        assert_eq!(
            array_op!(gen[16] | i | i),
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]
        );
        assert_eq!(array_op!(gen[4] | i | i as u8), [0u8, 1, 2, 3]);
        assert_eq!(
            array_op!(gen[16] | i | i as u8),
            [0u8, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]
        );

        const A: [u8; 4] = array_op!(gen[4] | i | i as u8);
        const B: [u8; 16] = array_op!(gen[16] | i | i as u8);
        assert_eq!(A, [0u8, 1, 2, 3]);
        assert_eq!(B, [0u8, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]);

        const LEN: usize = std::mem::size_of::<std::arch::x86_64::__m256i>();
        const INDICES: [u8; LEN] = array_op!(gen[LEN] | i | (i as u8) >> 1);
        assert_eq!(
            INDICES,
            [
                0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12,
                13, 13, 14, 14, 15, 15
            ]
        );
    }
}
